""" Car Battery implementation built upon the bslib library. It contains a CarBattery Class together with its Configuration and State. """

# clean

# Import packages from standard library or the environment e.g. pandas, numpy etc.
import importlib
from dataclasses import dataclass
from typing import Any, List

import pandas as pd
from bslib import bslib as bsl
from dataclasses_json import dataclass_json

# Import modules from HiSim
from hisim.component import (
    Component,
    ComponentConnection,
    ComponentInput,
    ComponentOutput,
    ConfigBase,
    SingleTimeStepValues,
    OpexCostDataClass,
    DisplayConfig,
    CapexCostDataClass
)
from hisim.loadtypes import ComponentType, InandOutputType, LoadTypes, Units
from hisim.simulationparameters import SimulationParameters
from hisim.postprocessing.kpi_computation.kpi_structure import KpiTagEnumClass, KpiEntry

__authors__ = "Tjarko Tjaden, Hauke Hoops, Kai Rösken"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = "..."
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Tjarko Tjaden"
__email__ = "tjarko.tjaden@hs-emden-leer.de"
__status__ = "development"


@dataclass_json
@dataclass
class CarBatteryConfig(ConfigBase):
    """Configuration of a Car Battery."""

    #: building_name in which component is
    building_name: str
    #: name of the device
    name: str
    #: priority of the device in hierachy: the higher the number the lower the priority
    source_weight: int
    #: name of battery to search in database (bslib)
    system_id: str
    #: charging and discharging power in Watt
    p_inv_custom: float
    #: battery capacity in in kWh
    e_bat_custom: float
    #: amount of energy used to charge the car battery
    charge: float
    #: amount of energy discharged from the battery
    discharge: float

    @classmethod
    def get_main_classname(cls):
        """Return the full class name of the base class."""
        return CarBattery.get_full_classname()

    @classmethod
    def get_default_config(cls, building_name: str = "BUI1", name: str = "CarBattery") -> "CarBatteryConfig":
        """Returns default configuration of a Car Battery."""
        config = CarBatteryConfig(
            building_name=building_name,
            name=name,
            system_id="SG1",
            p_inv_custom=1e4,
            e_bat_custom=30,
            source_weight=1,
            charge=0,
            discharge=0,
        )
        return config


class CarBattery(Component):
    """Car Battery class.

    Simulate state of charge and realized power of a ac coupled battery
    storage system with the bslib library. Relevant simulation parameters
    are loaded within the init for a specific or generic battery type.

    Components to connect to:
    (1) CarBattery controller (controller_l1_generic_ev_charge)
    """

    # Inputs
    LoadingPowerInput = "LoadingPowerInput"  # W

    # Outputs
    AcBatteryPower = "AcBatteryPower"  # W
    DcBatteryPower = "DcBatteryPower"  # W
    StateOfCharge = "StateOfCharge"  # [0..1]

    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: CarBatteryConfig,
        my_display_config: DisplayConfig = DisplayConfig(),
    ):
        """Loads the parameters of the specified battery storage."""
        self.battery_config = config
        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )

        self.source_weight = self.battery_config.source_weight

        self.system_id = self.battery_config.system_id

        self.p_inv_custom = self.battery_config.p_inv_custom

        self.e_bat_custom = self.battery_config.e_bat_custom

        # Component has states
        self.state = EVBatteryState(soc=0.5)
        self.previous_state = self.state.clone()

        # Load battery object with parameters from bslib database
        self.bat = bsl.ACBatMod(
            system_id=self.system_id,
            p_inv_custom=self.p_inv_custom,
            e_bat_custom=self.e_bat_custom,
        )

        # Define component inputs
        self.p_set: ComponentInput = self.add_input(
            object_name=self.component_name,
            field_name=self.LoadingPowerInput,
            load_type=LoadTypes.ELECTRICITY,
            unit=Units.WATT,
            mandatory=True,
        )

        # Define component outputs
        self.p_bs: ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.AcBatteryPower,
            load_type=LoadTypes.ELECTRICITY,
            unit=Units.WATT,
            postprocessing_flag=[
                InandOutputType.CHARGE_DISCHARGE,
                ComponentType.CAR_BATTERY,
            ],
            output_description="Charging power of the battery in Watt (Alternating current)",
        )

        self.p_bat: ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.DcBatteryPower,
            load_type=LoadTypes.ELECTRICITY,
            unit=Units.WATT,
            output_description="Charging power of the battery in Watt (Direct current).",
        )

        self.soc: ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.StateOfCharge,
            load_type=LoadTypes.ANY,
            unit=Units.ANY,
            postprocessing_flag=[InandOutputType.STORAGE_CONTENT],
            output_description="State of charge of the battery.",
        )

        self.add_default_connections(self.get_default_connections_from_charge_controller())

    def get_default_connections_from_charge_controller(self) -> Any:
        """Get default connections from charge controller."""
        # use importlib for importing the other component in order to avoid circular-import errors
        component_module_name = "hisim.components.controller_l1_generic_ev_charge"
        component_module = importlib.import_module(name=component_module_name)
        component_class = getattr(component_module, "L1Controller")
        connections: List[ComponentConnection] = []
        ev_charge_controller_classname = component_class.get_classname()
        connections.append(
            ComponentConnection(
                CarBattery.LoadingPowerInput,
                ev_charge_controller_classname,
                component_class.ToOrFromBattery,
            )
        )
        return connections

    def i_save_state(self) -> None:
        """Saves the state."""
        self.previous_state = self.state.clone()

    def i_restore_state(self) -> None:
        """Restores the state."""
        self.state = self.previous_state.clone()

    def i_doublecheck(self, timestep: int, stsv: SingleTimeStepValues) -> None:
        """Doublechecks."""
        pass

    def i_prepare_simulation(self) -> None:
        """Prepares the simulation."""
        pass

    def i_simulate(self, timestep: int, stsv: SingleTimeStepValues, force_convergence: bool) -> None:
        """Simulates the component."""
        # Parameters
        seconds_per_timestep = self.my_simulation_parameters.seconds_per_timestep

        # Load input values
        p_set = stsv.get_input_value(self.p_set)
        soc = self.state.soc

        # Simulate battery charging
        if p_set >= 0:
            results = self.bat.simulate(p_load=p_set, soc=soc, dt=seconds_per_timestep)
            p_bs = results[0]
            p_bat = results[1]
            soc = results[2]

        # Simulate battery discharge without losses (this is included in the car consumption of the car component)
        else:
            soc = soc + (p_set * self.my_simulation_parameters.seconds_per_timestep / 3600) / (self.e_bat_custom * 1e3)
            if soc < 0:
                raise ValueError(
                    "Car cannot drive, because battery is empty."
                    + "This points towards a major problem in the battery configuration - or the consumption pattern of the car."
                )
            p_bs = 0
            p_bat = 0

        # write values for output time series
        stsv.set_output_value(self.p_bs, p_bs)
        stsv.set_output_value(self.p_bat, p_bat)
        stsv.set_output_value(self.soc, soc)

        # write values to state
        self.state.soc = soc

    def write_to_report(self) -> List[str]:
        """Writes Car Battery values to report."""
        return self.battery_config.get_string_dict()

    def get_cost_opex(
        self,
        all_outputs: List,
        postprocessing_results: pd.DataFrame,
    ) -> OpexCostDataClass:
        """Calculate OPEX costs.

        No electricity costs for components except for Electricity Meter,
        because part of electricity consumption is feed by PV

        electricity_consumption is considered vor generic_car already
        """
        battery_losses_in_kwh: float = 0.0
        for index, output in enumerate(all_outputs):
            if (
                output.postprocessing_flag is not None
                and output.component_name == self.component_name
            ):
                if InandOutputType.CHARGE_DISCHARGE in output.postprocessing_flag:
                    self.battery_config.charge = round(
                        postprocessing_results.iloc[:, index].clip(lower=0).sum()
                        * self.my_simulation_parameters.seconds_per_timestep
                        / 3.6e6,
                        1,
                    )
                    self.battery_config.discharge = round(
                        postprocessing_results.iloc[:, index].clip(upper=0).sum()
                        * self.my_simulation_parameters.seconds_per_timestep
                        / 3.6e6,
                        1,
                    ) * (-1)
                    battery_losses_in_kwh = self.battery_config.charge - self.battery_config.discharge
        # Todo: Battery Aging like in component advanced_battery_bslib? Or is this considered in maintenance cost of car?

        opex_cost_data_class = OpexCostDataClass(
            opex_energy_cost_in_euro=0,
            opex_maintenance_cost_in_euro=0,
            co2_footprint_in_kg=0,
            consumption_in_kwh=battery_losses_in_kwh,
            loadtype=LoadTypes.ELECTRICITY,
            kpi_tag=KpiTagEnumClass.CAR_BATTERY
        )

        return opex_cost_data_class

    @staticmethod
    def get_cost_capex(config: CarBatteryConfig, simulation_parameters: SimulationParameters) -> CapexCostDataClass:  # pylint: disable=unused-argument
        """Returns investment cost, CO2 emissions and lifetime."""
        capex_cost_data_class = CapexCostDataClass.get_default_capex_cost_data_class()
        return capex_cost_data_class

    def get_component_kpi_entries(
        self,
        all_outputs: List,
        postprocessing_results: pd.DataFrame,
    ) -> List[KpiEntry]:
        """Calculates KPIs for the respective component and return all KPI entries as list."""
        return []


@dataclass
class EVBatteryState:
    """Electric vehicle battery state class."""

    # state of charge of the battery
    soc: float = 0

    def clone(self):
        """Creates a copy of the Car Battery State."""
        return EVBatteryState(soc=self.soc)



# clean

"""Postprocessing option computes overall consumption, production,self-consumption and injection as well as selfconsumption rate and self-sufficiency rate.

KPis for PV-battery systems in houses:
https://solar.htw-berlin.de/wp-content/uploads/WENIGER-2017-Vergleich-verschiedener-Kennzahlen-zur-Bewertung-von-PV-Batteriesystemen.pdf.
"""


from typing import List, Dict
from dataclasses import dataclass
from dataclass_wizard import JSONWizard
from hisim.loadtypes import DistrictNames
from hisim.postprocessing.postprocessing_datatransfer import PostProcessingDataTransfer
from hisim.postprocessing.kpi_computation.kpi_preparation import KpiPreparation, KpiTagEnumClass


@dataclass
class KpiGenerator(JSONWizard, KpiPreparation):
    """Class for generating and calculating key performance indicators."""

    post_processing_data_transfer: PostProcessingDataTransfer
    building_objects_in_district_list: list

    def __post_init__(self):
        """Build the dataclass from input data."""
        super().__init__(
            post_processing_data_transfer=self.post_processing_data_transfer,
            building_objects_in_district_list=self.building_objects_in_district_list,
        )

        for building_objects_in_district in self.building_objects_in_district_list:
            self.create_kpi_collection(building_objects_in_district)

        for building_objects_in_district in self.building_objects_in_district_list:
            if any(word in building_objects_in_district for word in DistrictNames):
                self.create_overall_district_kpi(district_name=building_objects_in_district)

        self.kpi_collection_dict_sorted = self.sort_kpi_collection_according_to_kpi_tags(
            kpi_collection_dict_unsorted=self.kpi_collection_dict_unsorted
        )

        self.return_table_for_report()

    def create_kpi_collection(self, building_objects_in_district):
        """Create kpi collection and write back into post processing data transfer."""
        # get filtered result dataframe
        self.filtered_result_dataframe = self.filter_results_according_to_postprocessing_flags(
            all_outputs=self.all_outputs,
            results=self.results,
            building_objects_in_district=building_objects_in_district,
        )

        # get consumption and production and battery kpis
        (
            total_electricity_consumption_in_kilowatt_hour,
            total_electricity_production_in_kilowatt_hour,
            pv_production_in_kilowatt_hour,
            windturbine_production_in_kilowatt_hour,
            building_production_in_kilowatt_hour,
        ) = self.compute_electricity_consumption_and_production_and_battery_kpis(
            result_dataframe=self.filtered_result_dataframe,
            building_objects_in_district=building_objects_in_district,
            kpi_tag=(
                KpiTagEnumClass.GENERAL
                if not any(word in building_objects_in_district for word in DistrictNames)
                else KpiTagEnumClass.ELECTRICITY_GRID
            ),
        )

        # get ratio between total production and total consumption
        self.compute_ratio_between_two_values_and_set_as_kpi(
            denominator_value=total_electricity_production_in_kilowatt_hour,
            numerator_value=total_electricity_consumption_in_kilowatt_hour,
            kpi_name="Ratio between total production and total consumption",
            building_objects_in_district=building_objects_in_district,
            kpi_tag=(
                KpiTagEnumClass.GENERAL
                if not any(word in building_objects_in_district for word in DistrictNames)
                else KpiTagEnumClass.ELECTRICITY_GRID
            ),
        )
        # get ratio between pv production and total consumption
        self.compute_ratio_between_two_values_and_set_as_kpi(
            denominator_value=pv_production_in_kilowatt_hour,
            numerator_value=total_electricity_consumption_in_kilowatt_hour,
            kpi_name="Ratio between PV production and total consumption",
            building_objects_in_district=building_objects_in_district,
            kpi_tag=(
                KpiTagEnumClass.GENERAL
                if not any(word in building_objects_in_district for word in DistrictNames)
                else KpiTagEnumClass.ELECTRICITY_GRID
            ),
        )
        # get ratio between wka production and total consumption
        self.compute_ratio_between_two_values_and_set_as_kpi(
            denominator_value=windturbine_production_in_kilowatt_hour,
            numerator_value=total_electricity_consumption_in_kilowatt_hour,
            kpi_name="Ratio between Windturbine production and total consumption",
            building_objects_in_district=building_objects_in_district,
            kpi_tag=(
                KpiTagEnumClass.GENERAL
                if not any(word in building_objects_in_district for word in DistrictNames)
                else KpiTagEnumClass.ELECTRICITY_GRID
            ),
        )
        # get ratio between building production and total consumption
        self.compute_ratio_between_two_values_and_set_as_kpi(
            denominator_value=building_production_in_kilowatt_hour,
            numerator_value=total_electricity_consumption_in_kilowatt_hour,
            kpi_name="Ratio between Building production and total consumption",
            building_objects_in_district=building_objects_in_district,
            kpi_tag=(
                KpiTagEnumClass.GENERAL
                if not any(word in building_objects_in_district for word in DistrictNames)
                else KpiTagEnumClass.ELECTRICITY_GRID
            ),
        )

        # get self-consumption, autarkie, injection
        self.filtered_result_dataframe = self.compute_self_consumption_injection_self_sufficiency(
            result_dataframe=self.filtered_result_dataframe,
            electricity_consumption_in_kilowatt_hour=total_electricity_consumption_in_kilowatt_hour,
            electricity_production_in_kilowatt_hour=total_electricity_production_in_kilowatt_hour,
            building_objects_in_district=building_objects_in_district,
            kpi_tag=(
                KpiTagEnumClass.GENERAL
                if not any(word in building_objects_in_district for word in DistrictNames)
                else KpiTagEnumClass.ELECTRICITY_GRID
            ),
        )
        # get electricity to and from grid
        (
            total_electricity_from_grid_in_kwh,
            total_electricity_to_grid_in_kwh,
        ) = self.get_electricity_to_and_from_grid_from_electricty_meter(building_objects_in_district)
        # get relative electricity demand
        relative_electricity_demand_from_grid_in_percent = self.compute_relative_electricity_demand(
            total_electricity_consumption_in_kilowatt_hour=total_electricity_consumption_in_kilowatt_hour,
            electricity_from_grid_in_kilowatt_hour=total_electricity_from_grid_in_kwh,
            building_objects_in_district=building_objects_in_district,
            kpi_tag=(
                KpiTagEnumClass.GENERAL
                if not any(word in building_objects_in_district for word in DistrictNames)
                else KpiTagEnumClass.ELECTRICITY_GRID
            ),
        )

        # get self-consumption rate according to solar htw berlin
        self.compute_self_consumption_rate_according_to_solar_htw_berlin(
            total_electricity_production_in_kilowatt_hour=total_electricity_production_in_kilowatt_hour,
            electricity_to_grid_in_kilowatt_hour=total_electricity_to_grid_in_kwh,
            building_objects_in_district=building_objects_in_district,
            kpi_tag=(
                KpiTagEnumClass.GENERAL
                if not any(word in building_objects_in_district for word in DistrictNames)
                else KpiTagEnumClass.ELECTRICITY_GRID
            ),
        )
        # get self-sufficiency rate according to solar htw berlin
        self.compute_self_sufficiency_according_to_solar_htw_berlin(
            relative_electricty_demand_in_percent=relative_electricity_demand_from_grid_in_percent,
            building_objects_in_district=building_objects_in_district,
            kpi_tag=(
                KpiTagEnumClass.GENERAL
                if not any(word in building_objects_in_district for word in DistrictNames)
                else KpiTagEnumClass.ELECTRICITY_GRID
            ),
        )

        # get capex and opex costs
        self.read_opex_and_capex_costs_from_results(building_object=building_objects_in_district)

    def create_overall_district_kpi(self, district_name):
        """Creation of overall district kpis."""

        (
            overall_production_district_in_kilowatt_hour,
            overall_consumption_district_in_kilowatt_hour,
            overall_self_consumption_district_in_kilowatt_hour,
        ) = self.create_overall_district_kpi_collection(district_name=district_name)

        self.create_overall_district_costs_collection(district_name=district_name)
        self.create_overall_district_emissions_collection(district_name=district_name)
        self.create_overall_district_contracting_collection(district_name=district_name, all_outputs=self.all_outputs)

        # get ratio between total production and total consumption
        self.compute_ratio_between_two_values_and_set_as_kpi(
            denominator_value=overall_production_district_in_kilowatt_hour,
            numerator_value=overall_consumption_district_in_kilowatt_hour,
            kpi_name="Overall ratio between total production and total consumption in district",
            building_objects_in_district=district_name,
            kpi_tag=KpiTagEnumClass.GENERAL,
        )

        # get electricity to and from grid
        (
            total_electricity_from_grid_in_kwh,
            total_electricity_to_grid_in_kwh,
        ) = self.get_electricity_to_and_from_grid_from_electricty_meter(district_name)
        # get relative electricity demand
        relative_electricity_demand_from_grid_in_percent = self.compute_relative_electricity_demand(
            total_electricity_consumption_in_kilowatt_hour=overall_consumption_district_in_kilowatt_hour,
            electricity_from_grid_in_kilowatt_hour=total_electricity_from_grid_in_kwh,
            building_objects_in_district=district_name,
            kpi_tag=KpiTagEnumClass.GENERAL,
            name="Overall relative electricity demand from grid in district",
        )

        self.compute_ratio_between_two_values_and_set_as_kpi(
            denominator_value=overall_self_consumption_district_in_kilowatt_hour,
            numerator_value=overall_consumption_district_in_kilowatt_hour,
            kpi_name="Overall self-sufficiency rate of electricity in district",
            building_objects_in_district=district_name,
            kpi_tag=KpiTagEnumClass.GENERAL,
        )

        self.compute_self_consumption_rate_according_to_solar_htw_berlin(
            total_electricity_production_in_kilowatt_hour=overall_production_district_in_kilowatt_hour,
            electricity_to_grid_in_kilowatt_hour=total_electricity_to_grid_in_kwh,
            building_objects_in_district=district_name,
            kpi_tag=KpiTagEnumClass.GENERAL,
            name="Overall self-consumption rate according to solar htw berlin in district",
        )

        self.compute_self_sufficiency_according_to_solar_htw_berlin(
            relative_electricty_demand_in_percent=relative_electricity_demand_from_grid_in_percent,
            building_objects_in_district=district_name,
            kpi_tag=KpiTagEnumClass.GENERAL,
            name="Overall self-sufficiency according to solar htw berlin in district",
        )

    def return_table_for_report(self):
        """Return a table with all kpis for the report."""
        table: List = []
        table.append(["Object", "KPI", "Value", "Unit"])
        for building_object in self.building_objects_in_district_list:
            for kpi_tag, kpi_entries in self.kpi_collection_dict_sorted[building_object].items():
                table.append([f"{building_object}", f"{kpi_tag}", "", ""])
                table.append(["--------", "--------------------", "", ""])
                for kpi_name, kpi_entry in kpi_entries.items():
                    value = kpi_entry["value"]
                    if value is not None:
                        value = round(value, 2)
                    unit = kpi_entry["unit"]
                    table.append([f"{building_object}", f"{kpi_name}: ", f"{value}", f"{unit}"])
                table.append(["\n", "\n", "\n", "\n"])

        #  Insert line break for report
        wrapped_table_data = []
        for row in table:
            wrapped_row = []
            for cell in row:
                wrapped_cell = "\n".join([cell[i : i + 80] for i in range(0, len(cell), 80)])
                wrapped_row.append(wrapped_cell)
            wrapped_table_data.append(wrapped_row)

        return wrapped_table_data

    def sort_kpi_collection_according_to_kpi_tags(self, kpi_collection_dict_unsorted: Dict) -> Dict:
        """Sort KPI collection dict according to KPI tags."""

        kpi_collection_dict_sorted: Dict[str, Dict] = {
            building_objects: {} for building_objects in self.building_objects_in_district_list
        }

        for building_object in self.building_objects_in_district_list:
            # get all tags and use as keys for sorted kpi dict
            for entry in kpi_collection_dict_unsorted[building_object].values():
                kpi_collection_dict_sorted[building_object].update({entry["tag"]: {}})

            # now sort kpi dict entries according to tags
            for kpi_name, entry in kpi_collection_dict_unsorted[building_object].items():
                for tag, tag_dict in kpi_collection_dict_sorted[building_object].items():
                    if entry["tag"] is tag:
                        tag_dict.update({kpi_name: entry})

        return kpi_collection_dict_sorted
